from __future__ import annotations

from pathlib import Path

from operator import attrgetter
import re
import sys
import sysconfig
import subprocess

from virtualenv.seed.wheels.util import Wheel

BUNDLE_SUPPORT = {
    "3.8": {
        "pip": "pip-25.0.1-py3-none-any.whl",
        "setuptools": "setuptools-75.3.2-py3-none-any.whl",
        "wheel": "wheel-0.45.1-py3-none-any.whl",
    },
    "3.9": {
        "pip": "pip-25.0.1-py3-none-any.whl",
        "setuptools": "setuptools-78.1.0-py3-none-any.whl",
        "wheel": "wheel-0.45.1-py3-none-any.whl",
    },
    "3.10": {
        "pip": "pip-25.0.1-py3-none-any.whl",
        "setuptools": "setuptools-78.1.0-py3-none-any.whl",
        "wheel": "wheel-0.45.1-py3-none-any.whl",
    },
    "3.11": {
        "pip": "pip-25.0.1-py3-none-any.whl",
        "setuptools": "setuptools-78.1.0-py3-none-any.whl",
        "wheel": "wheel-0.45.1-py3-none-any.whl",
    },
    "3.12": {
        "pip": "pip-25.0.1-py3-none-any.whl",
        "setuptools": "setuptools-78.1.0-py3-none-any.whl",
        "wheel": "wheel-0.45.1-py3-none-any.whl",
    },
    "3.13": {
        "pip": "pip-25.0.1-py3-none-any.whl",
        "setuptools": "setuptools-78.1.0-py3-none-any.whl",
        "wheel": "wheel-0.45.1-py3-none-any.whl",
    },
    "3.14": {
        "pip": "pip-25.0.1-py3-none-any.whl",
        "setuptools": "setuptools-78.1.0-py3-none-any.whl",
        "wheel": "wheel-0.45.1-py3-none-any.whl",
    },
}
MAX = "3.8"


def get_bundle_folder(for_py_version=None):
    """Return path of the system seed wheels"""
    global _BUNDLE_FOLDER_CACHE

    try:
        return _BUNDLE_FOLDER_CACHE[for_py_version]
    except KeyError:
        pass

    version_pattern = r"^\d+\.\d+$"
    if for_py_version and not re.match(version_pattern, for_py_version):
        raise ValueError(
            f"Unsupported value for for_py_version: '{for_py_version}', "
            f"expected regex: '{version_pattern}'"
        )
    if for_py_version is None:
        for_py_version_major = str(sys.version_info.major)
    else:
        for_py_version_major = for_py_version[0]

    # assume all system Pythons were configured with the same "scripts" path
    system_python = Path(sysconfig.get_path(name="scripts")) / (
        f"python{for_py_version_major}"
    )
    cmd = [
        system_python,
        "-c",
        (
            "import os, sys, system_seed_wheels;"
            "sys.stdout.write(os.path.dirname(system_seed_wheels.__file__))"
        ),
    ]
    try:
        _BUNDLE_FOLDER_CACHE[for_py_version] = Path(
            subprocess.check_output(cmd, text=True)
        )
        return _BUNDLE_FOLDER_CACHE[for_py_version]
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            "Cannot find the folder with system seed wheels. "
            "Please, install system-seed-wheels package if "
            f"it has support for your Python interpreter: '{system_python}'"
        ) from e


class BundleSupport:
    SEED_NAMES = ("pip", "setuptools", "wheel")

    def __init__(self, for_py_version=None):
        self._system_seed_wheels = None
        if for_py_version is None:
            self.for_py_version = "{}.{}".format(*sys.version_info[0:2])
        else:
            self.for_py_version = for_py_version
        self.wheels_dir = get_bundle_folder(for_py_version)

    def discover_wheel(self, distribution):
        """based on virtualenv.seed.wheels.util.discover_wheels"""
        wheels = []
        for filename in self.wheels_dir.iterdir():
            wheel = Wheel.from_path(filename)
            if wheel and wheel.distribution == distribution:
                wheels.append(wheel)
        sorted_wheels = sorted(
            wheels,
            key=attrgetter("version_tuple", "distribution"),
            reverse=True,
        )
        if not sorted_wheels:
            raise RuntimeError(
                f"Wheel for distribution: '{distribution}' not found on path: "
                f"'{self.wheels_dir}'"
            )
        return sorted_wheels[0].name

    @property
    def system_seed_wheels(self):
        if self._system_seed_wheels is None:
            wheels_names = {}
            for distr in self.SEED_NAMES:
                wheels_names[distr] = self.discover_wheel(distr)
            self._system_seed_wheels = {
                self.for_py_version: dict(wheels_names)
            }

        return self._system_seed_wheels

    def __getitem__(self, key):
        return self.system_seed_wheels[self.for_py_version]

    def get(self, key, default=None):
        return self.system_seed_wheels[self.for_py_version]

    def items(self):
        return self.system_seed_wheels.items()

    def keys(self):
        return self.system_seed_wheels.keys()


_BUNDLE_FOLDER_CACHE = {}
# defined only for self tests, which only check the current interpreter
BUNDLE_FOLDER = get_bundle_folder()
BUNDLE_SUPPORT = BundleSupport()


def get_embed_wheel(distribution, for_py_version):
    path = get_bundle_folder(for_py_version) / (
        BundleSupport(for_py_version).get(for_py_version, {})
    ).get(distribution)
    return Wheel.from_path(path)


__all__ = [
    "BUNDLE_FOLDER",
    "BUNDLE_SUPPORT",
    "MAX",
    "get_embed_wheel",
]
